############################################################
##
## The following package defines an iterator so that, given
## a context, it quickly iterates through all the child
## contexts.
##
############################################################

package Progeny;

sub new {
  my ($class, $top) = @_;
  my @gaps;
  my $i = $main::activeContexts >> 1;
  while ($i) {
    push @gaps, $i unless $i & $top;
    $i >>= 1;
  }
  bless [$top, 1 << @gaps, @gaps], $class;
}

sub progeny {
  my ($self) = @_;
  my $t = --($self->[1]);
  return 0 unless $t;
  my $i = 0;
  $t >>= 1, ++$i until $t & 1;
  my $flip = $self->[$i+2];
  my $p = $self->[0];
  $self->[0] = ( $p & $flip ? $p & ~$flip : $p | $flip);
}

############################################################
##
## The main program, based on Skousen's "new" algorithm.
##
############################################################

package main;

srand(time() ^ ($$ + ($$ << 15)));

readConfig();

if ($outputToFile eq 'yes') {
  open (PFLE, ">$resultsfile");
  select PFLE;
}

readData();
readOutcome();

readTestItems();

my $i;
for ($i = 0; $i < @testItems; $i++) {
  my ($sec,$min,$hour) = localtime();
  print "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "$i:  ";

  parse($testItems[$i]);
  print STDOUT "@{[@curTestItem]}\n";

  my $r;
  for ($r = 0; $r < $repeat; $r++) {
    setProb();
    printTop();
    initialize();
    sortbylevel();
    buildSet();
    countFreq();
    printResults();
  }
}

my ($sec,$min,$hour) = localtime();
print "Time: ${hour}h ${min}m ${sec}s\n\n";

if ($outputToFile eq 'yes') {
  close PFLE;
  print STDOUT "File is done.  Look at $resultsfile\n";
}

############################################################
sub readConfig {
  undef $/;
  open (CONFIG, "AM.config");
  $_ = <CONFIG>;
  close CONFIG;
  $/ = "\n";

  my ($name) = /Project Name.*:\s*(.*)/;
  $datafile = "$1.data";
  $outcomefile = "$1.outcome";
  $testfile = "$1.test";
  $resultsfile = "$1.results";

  ($excNull) = /null.*:\s*(.*)/;
  $excNull = 0 unless $excNull eq 'exclude';

  ($excGiven) = /given.*:\s*(.*)/;
  $excGiven = 0 unless $excGiven eq 'exclude';

  ($probability) = /probability.*:\s*(.*)/;
  ($repeat) = /repeat.*:\s*(.*)/;
  ($linearOrSq) = /linear or squared.*:\s*(.*)/;

  ($outputToFile) = /output to file.*:\s*(.*)/;

  ($useOutcomeFile) = /use \.outcome file.*:\s*(.*)/;

  ($specifyFreq) = /specify frequency.*:\s*(.*)/;
  $specifyFreq = 0 if $specifyFreq eq 'no';

  ($usescommas) = /format with commas.*:\s*(.*)/;
  if ($usescommas eq 'yes') {
    $bigsep = '\s*,\s*';           # assume comma-separated
    $smallsep = '\s+';             # and space-separated
  } else {
    $bigsep = '\s+';               # assume space-separated
    $smallsep = '';                # and not separated
  }
}

############################################################
sub readData {
  open(DATA, "$datafile");
  while(<DATA>) {
    chomp;
    my($outcome, $data, $spec) = split /$bigsep/, $_, 3;

    push @outcome, $outcome;
    push @spec, $spec;
    push @data, [split /$smallsep/, $data];

    last if $specifyFreq and ($specifyFreq == @data);

  }
  close DATA;
}

############################################################
sub readOutcome {
  undef $/;
  open(OUTCOME, "$outcomefile");
  $_ = <OUTCOME>;
  close OUTCOME;
  $/ = "\n";
  @outcomeIndex = split;
}

############################################################
sub readTestItems {
  open(TEST, "$testfile");
  while (<TEST>) {
    chomp;
    push @testItems, $_;
  }
  close TEST;
}

############################################################
sub parse {
  my ($outcome, $item);
  ($outcome, $item, $testspec) = split /$bigsep/, $_[0], 3;
  my $eq = ($excNull ? $item =~ tr/=/=/ : 0);
  @curTestItem = split /$smallsep/, $item;
  $activeVar = @curTestItem - $eq;
  $activeContexts = 1 << $activeVar;
}

############################################################
sub setProb {

  $excludedData = 0;
  $probvec = pack "b*", "1" x @data;

  return if $probability == 1;

  for (my $i=0; $i<@data; $i++) {
    vec($probvec, $i, 1) = 0, $excludedData++
      if rand() > $probability;
  }
}

############################################################
sub printTop {                     # much easier to use a
  print <<TOP;                     # here document
Project Name: $datafile
Given Context:  @curTestItem, $testspec
@{[ $excGiven ? "If context is in data file then exclude"
    : "Include context even if it is in the data file" ]}
Number of data points: @{[ scalar @data ]}
Probability of including any one data point: $probability
Total Excluded: $excludedData
Nulls: @{[ $excNull ? $excNull : "include" ]}
Gang: $linearOrSq
Number of active variables: $activeVar
Number of active contexts: $activeContexts

TOP
}

############################################################
sub initialize {
  $dead = pack "b*", "0" x $activeContexts;
  undef %lattice;
  undef %top;
}

############################################################
sub sortbylevel {
  undef @level;
  @contexts = 0 x @data;
  my $i;
  for ($i = 0; $i < @data; $i++) {
    next unless vec($probvec, $i, 1);
    my @dataItem = @{$data[$i]};
    my $l = 0;
    my $context = 0;
    my $j;
    for ($j = 0; $j < @curTestItem; $j++) {
      next if $curTestItem[$j] eq '=';
      my $bit = $curTestItem[$j] ne $dataItem[$j];
      $l =+ $bit;
      $contexts[$i] = ($contexts[$i] << 1) | $bit;
    }
    push @{$level[$l]}, $i;
  }
}

############################################################
sub buildSet {
  my $nullhashref = {};
  my $l;
  foreach $l (reverse(0..$activeVar-1),$activeVar) {
    my $i;
    foreach $i (@{$level[$l]}) {
      my $outcome = $outcome[$i];
      my $context = $contexts[$i];
      if (!$context) {
        print "Test item is in the data.\n";
        next if $excGiven;
      }
      next if vec($dead, $context, 1);

      undef %takencareof;

      my $ref;
      if (!($ref = $lattice{$context})) {
        ++$top{$context};
        $ref = $lattice{$context} = $nullhashref;
      }

      if ($top{$context} or
          (keys(%$ref) == 1 and exists $ref->{$outcome})) {
        my %h = clone($ref);
        push @{$h{$outcome}}, $i;
        $lattice{$context} = $ref
                           = $takencareof{$ref} = \%h;
      } else {
        vec($dead, $context, 1) = 1;
        delete $lattice{$context};
        undef $takencareof{$ref};
      }

      $ancestor = Progeny->new($context);
      my $p;
      while ($p = $ancestor->progeny()) {
        next if vec($dead, $p, 1);
        delete $top{$p};
        my $r;
        $lattice{$p} = $ref, next unless $r = $lattice{$p};
        if (exists $takencareof{$r}) {
          if (not defined $takencareof{$r}) {
            vec($dead, $p, 1) = 1;
            delete $lattice{$p};
          } else {
            $lattice{$p} = $takencareof{$r};
          }
          next;
        }
        if (keys(%$r) == 1 and exists $r->{$outcome}) {
          my %h = clone($r);
          push @{$h{$outcome}}, $i;
          $lattice{$p} = $takencareof{$r} = \%h;
        } else {
          vec($dead, $p, 1) = 1;
          delete $lattice{$p};
          undef $takencareof{$r};
        }
      }
    }
  }
}

############################################################
sub clone {
  my @array = %{shift(@_)};
  foreach (@array) {
    next unless ref($_);
    $_ = [ @$_ ];
  }
  return @array;
}

############################################################
sub countFreq {
  @datacount = %sum = ();
  $grandtotal = 0;
  my ($context, $pointer, %set, $n, @list);
  $linearOrSq eq 'linear' ? ($$n = 0) : ($n = \$#list);
  while (($context, $pointer) = each %lattice) {
    unless ($set{$pointer}) {
      my ($outcome, $listref);
      my @items = ();
      while (($outcome, $listref) = each %$pointer) {
        push @items, @$listref;
      }
      $set{$pointer} = \@items;
    }
    @list = @{$set{$pointer}};
    foreach (@list) {
      $datacount[$_] += 1 + $$n;
      $sum{$outcome[$_]} += 1 + $$n;
      $grandtotal += 1 + $$n;
    }
  }
}

############################################################
sub printResults {
  print "\n\nAnalogical Set\n\n";
  print "Total Frequency = $grandtotal\n\n";
  my $i;
  for ($i = 0; $i < @datacount; $i++) {
    next unless $datacount[$i];
    printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f\%\n",
      $outcome[$i],
      $spec[$i],
      $datacount[$i],
      100 * $datacount[$i]/$grandtotal;
  }
  print "\n\nStatistical Summary\n\n";
  for ($i = 0; $i < @outcomeIndex; $i += 2) {
    my $n;
    next unless $n = $sum{$outcomeIndex[$i]};
    printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f\%\n",
      $outcomeIndex[$i],
      $outcomeIndex[$i+1],
      $n,
      100 * $n/$grandtotal;
  }
  print "\n\n";
}

############################################################
##
##  Some subroutines useful for debugging.
##
############################################################

sub showlattice {
  my %l = %{shift(@_)};
  my ($a, $b);
  foreach $a (keys %l) {
    print "\n".dec2bin($a);
    foreach $b (keys %{$l{$a}}) {
      print "\t$b "
            .(join " ", map { $spec[$_] } @{$l{$a}->{$b}});
    }
  }
}

sub bin2dec {
  return unpack("N",
           pack("B32", substr("0" x 32 . shift, -32)));
}

sub dec2bin {
  my $str = unpack("B32", pack("N", shift));
  $str = substr($str, -$main::activeVar);
  return $str;
}
